from sonusai.utils.asr_functions.aixplain_whisper import aixplain_whisper
from sonusai.utils.asr_functions.deepgram import deepgram
from sonusai.utils.asr_functions.google import google
from sonusai.utils.asr_functions.whisper import whisper
